/**
   @author Shin'ichiro Nakaoka
*/

#ifndef CNOID_GUIBASE_PUT_PROPERTY_FUNCTION_H_INCLUDED
#define CNOID_GUIBASE_PUT_PROPERTY_FUNCTION_H_INCLUDED

#include <vector>
#include <string>
#include <boost/function.hpp>
#include "exportdecl.h"

namespace cnoid {

    class CNOID_EXPORT Selection
    {
    public:
        Selection(const char* domainname = 0)
            : domainname_(domainname) {
            selectedIndex_ = 0;
        }

        Selection(size_t size, const char* domainname = 0)
            : symbols_(size),
              domainname_(domainname) {
            selectedIndex_ = 0;
        }
        
        Selection& operator<<(const std::string& label){
            symbols_.push_back(label);
            return *this;
        }

        int selectedIndex() const {
            return selectedIndex_;
        }

        bool is(int index) const {
            return (index == selectedIndex_);
        }

        bool select(int index);
        bool select(const std::string& symbol);

        int size() const {
            return  symbols_.size();
        }
        
        void resize(int s) {
            symbols_.resize(s);
            if(selectedIndex_ >= s){
                selectedIndex_ = s - 1;
            }
        }

        const char* label(int index) const;
        const char* selectedLabel() const;

        void setSymbol(int index, const std::string& symbol){
            symbols_[index] = symbol;
        }

        std::string& symbol(int index) {
            return symbols_[index];
        }
        
        const std::string& symbol(int index) const {
            return symbols_[index];
        }
            
        const std::string& selectedSymbol() const {
            return symbols_[selectedIndex_];
        }

    private:
        std::vector<std::string> symbols_;
        int selectedIndex_;
        const char* domainname_;
    };

    class PutPropertyFunction
    {
    public:
        
        virtual ~PutPropertyFunction() { }

        virtual PutPropertyFunction& decimals(int d) = 0;
        virtual PutPropertyFunction& min(double min) = 0;
        virtual PutPropertyFunction& max(double max) = 0;
        virtual PutPropertyFunction& min(int min) = 0;
        virtual PutPropertyFunction& max(int max) = 0;
        virtual PutPropertyFunction& reset() = 0;

        // bool
        virtual void operator()(const std::string& name, bool value) = 0;
        virtual void operator()(const std::string& name, bool value, const boost::function<bool(bool)>& changeFunc) = 0;

        // int
        virtual void operator()(const std::string& name, int value) = 0;
        virtual void operator()(const std::string& name, int value, const boost::function<bool(int)>& changeFunc) = 0;

        // double
        virtual void operator()(const std::string& name, double value) = 0;
        virtual void operator()(const std::string& name, double value,
                                const boost::function<bool(double)>& changeFunc) = 0;

        // string
        virtual void operator()(const std::string& name, const std::string& value) = 0;
        virtual void operator()(const std::string& name, const std::string& value,
                                const boost::function<bool(const std::string&)>& changeFunc) = 0;

        // selection
        virtual void operator()(const std::string& name, const Selection& selection) = 0;
        virtual void operator()(const std::string& name, const Selection& selection,
                                const boost::function<bool(int which)>& changeFunc) = 0;
    };


    template <class ValueType>
    class ChangeProperty
    {
        ValueType& variable;
    public:
        ChangeProperty(ValueType& variable) : variable(variable) { }
        bool operator()(const ValueType& value){
            variable = value;
            return true;
        }
    };

    template <>
    class ChangeProperty<Selection>
    {
        Selection& selection;
    public:
        ChangeProperty(Selection& variable) : selection(variable) { }
        bool operator()(int value){
            selection.select(value);
            return true;
        }
    };
    
    template<class ValueType>
    ChangeProperty<ValueType> changeProperty(ValueType& variable) {
        return ChangeProperty<ValueType>(variable);
    }

}

#endif
