/**
   \file
   \author Shin'ichiro Nakaoka
*/

#include "BodyMotionEngine.h"
#include "BodyItem.h"
#include "BodyMotionItem.h"
#include <cnoid/ExtensionManager>
#include <boost/bind.hpp>

using namespace boost;
using namespace cnoid;

namespace {

    const bool TRACE_FUNCTIONS = false;
}

namespace cnoid {

    class BodyMotionEngineImpl
    {
    public:
        BodyMotionEngine* self;
        BodyItemPtr bodyItem;
        BodyPtr body;
        MultiValueSeqPtr qSeq;
        MultiSE3SeqPtr positions;
        Vector3SeqPtr ZMPseq;
        bool calcForwardKinematics;

        BodyMotionEngineImpl(BodyMotionEngine* self)
            : self(self) {

        }

        bool setMotionItem(Item* item) {

            BodyMotionItem* bodyMotionItem;
            MultiValueSeqItem* multiValueSeqItem;
            MultiSE3SeqItem* multiSE3SeqItem;;

            if((bodyMotionItem = dynamic_cast<BodyMotionItem*>(item))){
                if(setBodyItem(item)){
                    qSeq = bodyMotionItem->jointPosSeq();
                    positions = bodyMotionItem->linkPosSeq();
                    if(bodyMotionItem->hasZMPseqItem()){
                        ZMPseq = bodyMotionItem->ZMPseq();
                    }
                }

            } else if((multiValueSeqItem = dynamic_cast<MultiValueSeqItem*>(item))){
                if(setBodyItem(item)){
                    qSeq = multiValueSeqItem->seq();
                }
                
            } else if((multiSE3SeqItem = dynamic_cast<MultiSE3SeqItem*>(item))){
                if(setBodyItem(item)){
                    positions = multiSE3SeqItem->seq();
                }
            }

            if(bodyItem){
                calcForwardKinematics = !(positions && positions->numParts() > 1);
                item->sigUpdated().connect(bind(&BodyMotionEngine::notifyUpdate, self));
            }
            
            return (bodyItem);
        }

        bool setBodyItem(Item* item){
            bodyItem = item->findOwnerItem<BodyItem>();
            if(bodyItem){
                body = bodyItem->body();
            }
            return (bodyItem);
        }
        
        virtual bool onTimeChanged(double time){

            bool isActive = false;
            bool fkDone = false;
            
            if(qSeq){
                isActive = setJointPositions(time);
            }
            if(positions){
                isActive |= setLinkPositions(time);
                if(positions->numParts() == 1){
                    body->calcForwardKinematics(); // FK from the root
                    fkDone = true;
                }
            }
            if(ZMPseq){
                if(calcForwardKinematics){
                    bodyItem->calcForwardKinematics();
                    fkDone = true;
                }
                isActive |= setZMP(time);
            }

            bodyItem->notifyKinematicStateChange(!fkDone && calcForwardKinematics);

            return isActive;
        }

        static inline bool clamp(int& frame, int numFrames) {

            if(frame < 0){
                frame = 0;
                return false;
            } else if(frame >= numFrames){
                frame = numFrames - 1;
                return false;
            }
            return true;
        }

        bool setLinkPositions(double time){

            bool isValidTime = false;

            const int numLinks = positions->numParts();
            const int numFrames = positions->numFrames();
            if(numLinks > 0 && numFrames > 0){
                int frame = positions->frameOfTime(time);
                isValidTime = clamp(frame, numFrames);
                for(int i=0; i < numLinks; ++i){
                    Link* link = body->link(i);
                    const SE3& position = positions->at(frame, i);
                    link->p = position.translation();
                    link->R = position.rotation();
                }
            }

            return isValidTime;
        }

        bool setJointPositions(double time){

            bool isValidTime = false;

            const int numJoints = std::min(body->numJoints(), qSeq->numParts());
            int frame = qSeq->frameOfTime(time);
            const int numFrames = qSeq->numFrames();
            if(numJoints > 0 && numFrames > 0){
                isValidTime = clamp(frame, numFrames);
                MultiValueSeq::Frame q = qSeq->frame(frame);
                for(int i=0; i < numJoints; ++i){
                    body->joint(i)->q = q[i];
                }
            }

            return isValidTime;
        }    

        bool setZMP(double time){

            bool isValidTime = false;
            int frame = ZMPseq->frameOfTime(time);
            const int numFrames = ZMPseq->numFrames();
            if(numFrames > 0){
                isValidTime = clamp(frame, numFrames);
                bodyItem->setZmp((*ZMPseq)[frame]);
            }
            return isValidTime;
        }    

    };

    TimeSyncItemEnginePtr createBodyMotionEngine(Item* sourceItem)
    {
        BodyMotionEnginePtr engine = new BodyMotionEngine();
        if(!engine->setMotionItem(sourceItem)){
            engine = 0;
        }
        return engine;
    }
}


BodyMotionEngine::BodyMotionEngine()
{
    impl = new BodyMotionEngineImpl(this);
}


BodyMotionEngine::~BodyMotionEngine()
{
    delete impl;
}


bool BodyMotionEngine::setMotionItem(Item* item)
{
    return impl->setMotionItem(item);
}


bool BodyMotionEngine::onTimeChanged(double time)
{
    return impl->onTimeChanged(time);
}


void BodyMotionEngine::initialize(ExtensionManager* ext)
{
    ext->timeSyncItemEngineManger().addEngineFactory(createBodyMotionEngine);
}
